unit MainForm;

interface

uses Windows, Messages, NForm, NCommand, NkVImage, ImgFilter, SpiFilter,
  InfoDialog, ContextMenu, FileList;

type
  TNkVMainForm = class(TForm)
    Command: TCommand;
    Image: TNkVImage;
    InfoDlg: TInfoDlg;
    FileList: TFileList;
    FileListUpdate: Boolean;
    Archiv: TArchive;
    ArFileList: TFileList;
    DirMode: Boolean;
    FitImage: Boolean;
    FitLargeImage: Boolean;
    DispWidth, DispHeight: Integer;
    ViewLeft, ViewTop: Integer;
    ViewWidth, ViewHeight: Integer;
    OrgLeft, OrgTop: Integer;
    ShowGrid: Boolean;
    ScaleFilter: Boolean;
    Plugins: array of string;
    ArrowCursor: HCURSOR;
    HandCursor: HCURSOR;
    GrabCursor: HCURSOR;
    OldPt: TPoint;
    ShowColorValue: Boolean;
    ZoomValue: Double;
    BgColor: COLORREF;
    ContextMenu: TContextMenu;
    constructor Create(const Caption: string; ALeft, ATop, AWidth,
      AHeight: Integer); override;
    destructor Destroy; override;
    procedure SavePos;
    procedure Zoom(Value: Double);
    procedure Scroll(X, Y: Integer);
    procedure LoadConfig;
    procedure LoadPluginList;
    procedure LoadKeyMap;
    function LookupFilter(Buf: PChar; BufSize: Integer): TObject;
    procedure LoadFirstFile(const List: TFileList);
    procedure LoadFile(const Path: string; Last: Boolean);
    procedure LoadFirstFileFromArchive(const Ar: TArchive;
      const List: TFileList);
    procedure LoadLastFileFromArchive(const Ar: TArchive;
      const List: TFileList);
    procedure LoadFileFromArchive(const Ar: TArchive; const FileName: string);
    procedure SetClipboardText(const S: string);
    procedure NextImage(const List: TFileList);
    procedure PrevImage(const List: TFileList; Last: Boolean);
    procedure GetImageCoord(Cx, Cy: Integer; var X, Y: Integer);
    procedure GetColorValue(Ix, Iy: Integer; var R, G, B: Integer);
    procedure UpdateDispSize;
    procedure UpdateCaption;
    procedure UpdateCursor;
    procedure UpdateInfo;
    procedure UpdateWinSize;
    procedure FillFileList;
    procedure Error(const Path: string; const Msg: string);
    procedure SetWallpaper(Tile: Boolean);
    procedure ShowPluginConfig(const FileName: string);
    procedure ActivateNextForm(Prev: Boolean);
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMLButtonDown(var Message: TWMLButtonDown);message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk);
      message WM_LBUTTONDBLCLK;
    procedure WMContextMenu(var Message: TMessage); message WM_CONTEXTMENU;
    procedure WMInitMenuPopup(var Message: TMessage); message WM_INITMENUPOPUP;
    procedure WMDrawItem(var Message: TMessage); message WM_DRAWITEM;
    procedure WMMeasureItem(var Message: TMessage); message WM_MEASUREITEM;
    procedure WMDropFiles(var Message: TWMDropFiles); message WM_DROPFILES;
    procedure WMMouseWheel(var Message: TMessage); message WM_MOUSEWHEEL;
    procedure WMClose(var Message: TWMClose); message WM_CLOSE;
    procedure WMQueryEndSession(var Message: TWMQueryEndSession);
        message WM_QUERYENDSESSION;
  published
    procedure cmAbout;
    procedure cmAboutPlugins;
    procedure cmCopyColorValue;
    procedure cmCopyFileName;
    procedure cmExit;
    procedure cmFirstFile;
    procedure cmFitImage;
    procedure cmFitLargeImage;
    procedure cmFitWindow;
    procedure cmFullScreen;
    procedure cmHideBorder;
    procedure cmInfo;
    procedure cmLastFile;
    procedure cmNextFile;
    procedure cmNextImage;
    procedure cmNextWindow;
    procedure cmPrevFile;
    procedure cmPrevImage;
    procedure cmPrevWindow;
    procedure cmRefreshFileList;
    procedure cmReloadImage;
    procedure cmSaveAs;
    procedure cmScaleFilter;
    procedure cmScrollBottom;
    procedure cmScrollDown;
    procedure cmScrollLeft;
    procedure cmScrollRight;
    procedure cmScrollTop;
    procedure cmScrollTopBottom;
    procedure cmScrollUp;
    procedure cmSetImageSize;
    procedure cmSetWallpaper;
    procedure cmSetWallpaperTile;
    procedure cmShowColorValue;
    procedure cmShowGrid;
    procedure cmStayOnTop;
    procedure cmZoom100;
    procedure cmZoom150;
    procedure cmZoom200;
    procedure cmZoom300;
    procedure cmZoom400;
    procedure cmZoom50;
    procedure cmZoom800;
    procedure cmZoomIn;
    procedure cmZoomOut;
    procedure cmPluginMenu;
  end;

implementation

uses ShellAPI, NTypes, NLib, NConf, ifdib, ifpnm, keymenu, ImageSizeDialog;

{$INCLUDE version.inc}

const
  ZoomTable: array [0..18] of Double = (
    0.0833, 0.125, 0.167, 0.25, 0.333, 0.5, 0.667,
    1.0, 1.25, 1.5, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 12.0, 16.0
  );
  KeyScroll = 16;
  IDM_PLUGINSETTING_FIRST = 4000;
  IDM_PLUGINSETTING_LAST = IDM_PLUGINSETTING_FIRST + 999;

function TNkVMainForm.LookupFilter(Buf: PChar; BufSize: Integer): TObject;
const
  Filters: array[0..1] of TImgFilterClass = (TDibFilter, TPnmFilter);
var
  i: Integer;
begin
  for i := 0 to High(Filters) do
  begin
    Result := Filters[i].Create;
    if (Result as TImgFilter).Test(Buf, BufSize) then
      Exit
    else
      Result.Free;
  end;

  for i := 0 to High(Plugins) do
  begin
    Result := TSpiFilter.CreateByFile(Plugins[i]);
    if (Result as TImgFilter).Test(Buf, BufSize) then
    begin
      if (Result is TSpiFilter) and (Result as TSpiFilter).IsArchive then
      begin
        Result.Free;
        Result := TArchive.Create(Plugins[i]);
      end;
      Exit
    end else
      Result.Free;
  end;

  Result := nil;
end;

constructor TNkVMainForm.Create(const Caption: string; ALeft, ATop, AWidth,
  AHeight: Integer);
begin
  inherited;
  ZoomValue := 1.0;
  BgColor := GetSysColor(COLOR_WINDOW);
  ArrowCursor := LoadCursor(0, IDC_ARROW);
  HandCursor := LoadCursor(HInstance, 'HANDOPEN');
  GrabCursor := LoadCursor(HInstance, 'HANDHOLD');
  DragAcceptFiles(Handle, True);
  SetBackground(GetStockObject(NULL_BRUSH));

  Command := TCommand.Create(Self);
  Command.BindKeys(DefaultKeys);
  LoadKeyMap;
  App.Accel := Command.AccelHandle;

  LoadPluginList;
  Image := TNkVImage.Create;
  InfoDlg := TInfoDlg.CreateByResource(Handle, 'INFODIALOG');
  FileList := TFileList.Create;
  ArFileList := TFileList.Create;
end;

destructor TNkVMainForm.Destroy;
begin
  Command.Free;
  FileList.Free;
  ArFileList.Free;
  InfoDlg.Free;
end;

procedure TNkVMainForm.SavePos;
begin
  try
    SaveFormPos(JoinPath([App.ProfDir, 'position']), Self);
  except
    on E: EIOError do App.Error(E.Path+EOL+EOL+E.Msg);
  end;
end;

procedure TNkVMainForm.LoadPluginList;
var
  i: Integer;
  tail: string;
  dir: TDir;
begin
  SetLength(Plugins, 32);
  i := 0;
  try
    with TConf.Create(JoinPath([App.ProfDir, 'pluginlist'])) do
      try
        while ReadLine do
        begin
          if i = Length(Plugins) then
            SetLength(Plugins, i + i div 4);
          Plugins[i] := Line;
          Inc(i);
        end;
      finally
        Free;
      end;
  except
    on E: EIOError do
      with TListDir.Create(DirName(ProgPath)) do
        try
          while Each(dir) do
          begin
            if dmDirectory in dir.Mode then
              Continue;
            tail := Copy(dir.Name, Length(dir.Name)-3, 4);
            if (tail = '.spi') or (tail = '.SPI') then
            begin
              if i = Length(Plugins) then
                SetLength(Plugins, i + i div 4);
              Plugins[i] := dir.Name;
              Inc(i);
            end;
          end;
        finally
          Free;
        end;
  end;
  SetLength(Plugins, i);
end;

procedure TNkVMainForm.LoadConfig;
  function ToBool(const S: string): Boolean;
  begin
    Result := S = 'on';
  end;
var
  s: string;
  w, h, r, g, b: Integer;
begin
  w := 320;
  h := 240;
  try
    with TConf.Create(JoinPath([App.ProfDir, 'config'])) do
      try
        while ReadLine do
        begin
          s := GetArg(0);
          if s = 'fitimage' then
          begin
            FitImage := ToBool(GetArg(1));
            if FitImage then
              FitLargeImage := False;
          end
          else if s = 'fitlargeimage' then
          begin
            FitLargeImage := ToBool(GetArg(1));
            if FitLargeImage then
              FitImage := False;
          end
          else if s = 'scalefilter' then
            ScaleFilter := ToBool(GetArg(1))
          else if s = 'width' then
            w := SInt(GetArg(1))
          else if s = 'height' then
            h := SInt(GetArg(1))
          else if s = 'background' then
          begin
            r := SInt(GetArg(1));
            g := SInt(GetArg(2));
            b := SInt(GetArg(3));
            BGColor := RGB(r, g, b);
          end
          else if s = 'hideborder' then
            SetBorder(not ToBool(GetArg(1))); 
        end;
      finally
        Free;
      end;
  except
    on E: EIOError do ;
  end;
  SetSize(w, h);
end;

procedure TNkVMainForm.LoadKeyMap;
var
  i: Integer;
  s: array of string;
begin
  SetLength(s, 256);
  for i := 0 to High(s) do
    s[i] := '';
  i := 0;
  try
    with TConf.Create(JoinPath([App.ProfDir, 'keymap'])) do
      try
        while ReadLine and (i < High(s)) do
        begin
          s[i] := GetArg(0);;
          s[i+1] := GetArg(1);
          Inc(i, 2);
        end;
      finally
        Free;
      end;
  except
    on E: EIOError do ;
  end;
  SetLength(s, i);
  Command.BindKeys(s);
end;

procedure TNkVMainForm.FillFileList;
var
  s, path: string;
  i: Integer;
  dir: TDir;
begin
  if not DirMode then
    Exit;

  if not FileListUpdate then
    Exit;

  FileListUpdate := False;
  s := FileList.Items[FileList.Cur];
  path := DirName(s);
  FileList.Clear;
  with TListDir.Create(path) do
    try
      while Each(dir) do
        if not (dmDirectory in dir.Mode) then
          FileList.Add(JoinPath([path, dir.Name]));
    finally
      Free;
    end;
  FileList.Sort;
  
  for i := 0 to FileList.Count-1 do 
    if FileList.Items[i] = s then
    begin
      FileList.Cur := i;
      Exit;
    end;
end;

procedure TNkVMainForm.LoadFirstFile(const List: TFileList);
var
  i: Integer;
  multi, stdin: Boolean;
begin
  multi := List.Count > 1;
  stdin := (List.Count > 0) and (List.Items[0] = '');
  while List.Count > 0 do
  begin
    try
      LoadFile(List.Items[0], False)
    except
      on E: EIOError do
      begin
        if not stdin then
          Error(E.Path, E.Msg);
        List.Delete(0);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        if not multi then
          Error(List.Items[0], E.Msg);
        List.Delete(0);
        Continue;
      end;
    end;
    Break;
  end;

  if List.Count > 0 then // loding succeeded
  begin
    FileList.Clear;
    for i := 0 to List.Count - 1 do
      FileList.Add(List.Items[i]);
    FileList.Cur := 0;
    ViewLeft := 0;
    ViewTop := 0;
    DispWidth := RoundOff(Image.Width * ZoomValue);
    DispHeight := RoundOff(Image.Height * ZoomValue);
    UpdateDispSize;
    UpdateWinSize;
    UpdateInfo;
    UpdateCursor;
  end
  else if not stdin and multi then
    Error('', 'Loadable file not found.');
end;

procedure TNkVMainForm.UpdateCursor;
var
  cursor: HCURSOR;
begin
  if ShowColorValue then
    cursor := LoadCursor(0, IDC_CROSS)
  else if Border then
    cursor := HandCursor
  else
    cursor := ArrowCursor;
  SetClassLong(Handle, GCL_HCURSOR, cursor);
  SetCursor(cursor);
end;

procedure TNkVMainForm.LoadFileFromArchive(const Ar: TArchive;
  const FileName: string);
const
  TestBufSize = 2048;
var
  buf: PChar;
  fsize: Cardinal;
  filter: TObject;
  oldcur: HCURSOR;
begin
  oldcur := SetCursor(LoadCursor(0, IDC_WAIT));
  fsize := Ar[FileName].FileSize;
  buf := nil;
  try
    GetMem(buf, fsize+TestBufSize);
    FillChar(buf^, TestBufSize, 0);
    Ar.Extract(FileName, buf, fsize);
    filter := LookupFilter(buf, fsize);
    try
      if not Assigned(filter) or (filter is TArchive) then
        raise EImgFilterError.Create('Unknown format.');
      Image.ReadMem(buf, fsize, filter as TImgFilter);
    finally
      filter.Free;
    end;
  finally
    FreeMem(buf);
    SetCursor(oldcur);
  end;
end;

procedure TNkVMainForm.UpdateWinSize;
begin
  if not FitImage and not FitLargeImage and not FullScreen and
      not Maximized then
    SetClientSize(DispWidth, DispHeight);
  Update;
end;

procedure TNkVMainForm.LoadFirstFileFromArchive(const Ar: TArchive;
  const List: TFileList);
begin
  while List.Count > 0 do
  begin
    try
      LoadFileFromArchive(Ar, List.Items[0]);
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        List.Delete(0);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        //Error(List[0], E.Msg);
        List.Delete(0);
        Continue;
      end;
    end;
    Break;
  end;
  if List.Count > 0 then // loding succeeded
  begin
    Archiv.Free;
    Archiv := Ar;
    ArFileList.Free;
    ArFileList := List;
    ArFileList.Cur := 0;
  end else
    raise EImgFilterError.Create('Loadable file not found in archive.');
end;

procedure TNkVMainForm.LoadLastFileFromArchive(const Ar: TArchive;
  const List: TFileList);
begin
  while List.Count > 0 do
  begin
    try
      LoadFileFromArchive(Ar, List.Items[List.Count-1])
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        List.Delete(List.Count-1);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        //Error(List[List.Count-1], E.Msg);
        List.Delete(List.Count-1);
        Continue;
      end;
    end;
    Break;
  end;
  if List.Count > 0 then // loding succeeded
  begin
    Archiv.Free;
    Archiv := Ar;
    ArFileList.Free;
    ArFileList := List;
    ArFileList.Cur := List.Count-1;
  end else
    raise EImgFilterError.Create('Loadable file not found in archive.');
end;

procedure TNkVMainForm.LoadFile(const Path: string; Last: Boolean);
const
  TestBufSize = 2048;
  StdinBufSize = 32768;
var
  len, fsize: Cardinal;
  buf: PChar;
  filter: TObject;
  f: TFile;
  isstdin: Boolean;
  oldcur: HCURSOR;
  ar: TArchive;
  arlist: TFileList;
  fname: string;
begin
  isstdin := Path = '';
  if isstdin then
    f := TFile.CreateStdIn
  else
    f := TFile.Open(Path);

  buf := nil;
  oldcur := SetCursor(LoadCursor(0, IDC_WAIT));
  with f do
    try
      GetMem(buf, TestBufSize);
      FillChar(buf^, TestBufSize, 0);
      len := Read(buf^, TestBufSize);
      if isstdin and (len = 0) then
        raise EIOError.Create('<STDIN>');
      filter := LookupFilter(buf, len);
      try
        if not Assigned(filter) then
          raise EImgFilterError.Create('Unknown format.');

        if filter is TArchive then
        begin
          ar := filter as TArchive;
          ar.Open(Path);
          arlist := TFileList.Create;
          while ar.EachName(fname) do
            if ar.Info[fname].FileSize > 0 then
              arlist.Add(fname);
          arlist.Sort;
          if Last then
            LoadLastFileFromArchive(ar, arlist)
          else
            LoadFirstFileFromArchive(ar, arlist);
        end
        else begin
          if isstdin then
          begin
            // stdin
            fsize := len;
            while True do
            begin
              ReallocMem(buf, fsize + StdinBufSize);
              try
                len := Read(buf[fsize], StdinBufSize);
              except
                on E: EIOError do Break;
              end;
              if len = 0 then
                Break;
              Inc(fsize, len);
            end;
          end else
          begin
            // normal file
            fsize := Size;
            ReallocMem(buf, fsize);
            Read(buf[len], fsize - len);
          end; 
          Image.ReadMem(buf, fsize, filter as TImgFilter);
          ArFileList.Clear;
        end;
      finally
        if not (filter is TArchive) then
          filter.Free;
      end;
    finally
      FreeMem(buf);
      Free;
      SetCursor(oldcur);
    end;
end;

procedure TNkVMainForm.Zoom(Value: Double);
begin
  if Image.Loaded then
  begin
    FitImage := False;
    FitLargeImage := False;
    DispWidth := RoundOff(Image.Width * Value);
    DispHeight := RoundOff(Image.Height * Value);
    UpdateDispSize;
  end;
end;

procedure TNkVMainForm.Scroll(X, Y: Integer);
var
  cw, ch, vw, vh: Integer;
begin
  if not Image.Loaded then
    Exit;

  if FitImage or FitLargeImage then
    Exit;

  GetClientSize(cw, ch);
  if DispWidth < cw then
    vw := DispWidth
  else
    vw := cw;
  if DispHeight < ch then
    vh := DispHeight
  else
    vh := ch;

  Inc(ViewLeft, X);
  if ViewLeft < 0 then
  begin
    Dec(X, ViewLeft);
    ViewLeft := 0;
  end
  else if ViewLeft > (DispWidth-vw) then
  begin
    Inc(X, DispWidth-vw-ViewLeft);
    ViewLeft := DispWidth-vw;
  end;

  Inc(ViewTop, Y);
  if ViewTop < 0 then
  begin
    Dec(Y, ViewTop);
    ViewTop := 0;
  end
  else if ViewTop > (DispHeight-vh) then
  begin
    Inc(Y, DispHeight-vh-ViewTop);
    ViewTop := DispHeight-vh;
  end;

  ScrollWindowEx(Handle, -X, -Y, nil, nil, 0, nil, SW_INVALIDATE);
  Update;
end;

procedure TNkVMainForm.SetClipboardText(const S: string);
var
  h: HGLOBAL;
begin
  h := GlobalAlloc(GHND, Length(S)+1);
  Move(S[1], PChar(GlobalLock(h))^, Length(S));
  GlobalUnlock(h);
  if OpenClipboard(Handle) then
  begin
    EmptyClipboard;
    SetClipboardData(CF_TEXT, h);
    CloseClipBoard;
  end else
    GlobalFree(h);
end;

procedure TNkVMainForm.NextImage(const List: TFileList);
var
  i: Integer;
begin
  i := list.Cur+1;
  while i < list.Count do
  begin
    try
      if list = ArFileList then
        LoadFileFromArchive(Archiv, list.Items[i])
      else
        LoadFile(list.Items[i], False)
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        list.Delete(i);
        Continue;
      end;
      on E: EImgFilterError do // XXX
      begin
        list.Delete(i);
        Continue;
      end;
    end;
    Break;
  end;
  if i < list.Count then
    list.Cur := i;
  ViewLeft := 0;
  ViewTop := 0;
  DispWidth := RoundOff(Image.Width * ZoomValue);
  DispHeight := RoundOff(Image.Height * ZoomValue);
  UpdateDispSize;
  UpdateWinSize;
  UpdateInfo;
end;

procedure TNkVMainForm.PrevImage(const List: TFileList; Last: Boolean);
var
  i: Integer;
begin
  i := list.Cur-1;
  while i >= 0 do
  begin
    try
      if list = ArFileList then
        LoadFileFromArchive(Archiv, list.Items[i])
      else
        LoadFile(list.Items[i], Last)
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        FileList.Delete(i);
        Dec(i);
        Dec(FileList.Cur);
        Continue;
      end;
      on E: EImgFilterError do // XXX
      begin
        list.Delete(i);
        Dec(i);
        Dec(list.Cur);
        Continue;
      end;
    end;
    Break;
  end;
  if i >= 0 then
    list.Cur := i;
  ViewLeft := 0;
  ViewTop := 0;
  DispWidth := RoundOff(Image.Width * ZoomValue);
  DispHeight := RoundOff(Image.Height * ZoomValue);
  UpdateDispSize;
  UpdateWinSize;
  UpdateInfo;
end;

procedure TNkVMainForm.UpdateInfo;
var
  dir: TDir;
begin
  InfoDlg.Path := AbsPath(FileList.Items[FileList.Cur]);
  if ArFileList.Count = 0 then
  begin
    InfoDlg.PathInArchive := '';
    if InfoDlg.Path <> '' then
    begin
      dir := DirStat(InfoDlg.Path);
      InfoDlg.MTime := dir.MTime;
      InfoDlg.ATime := dir.ATime;
    end;
  end
  else begin
    InfoDlg.PathInArchive := ArFileList.Items[ArFileList.Cur];
    InfoDlg.MTime := Archiv[InfoDlg.PathInArchive].DateTime;
    FillChar(InfoDlg.ATime, SizeOf(InfoDlg.Atime), 0);
  end;
  InfoDlg.Image := Image;

  if IsWindow(InfoDlg.Handle) then
    InfoDlg.UpdateInfo;
end;

procedure TNkVMainForm.GetImageCoord(Cx, Cy: Integer; var X, Y: Integer);
begin
  X := Trunc((Cx + ViewLeft-OrgLeft) / ZoomValue);
  Y := Trunc((Cy + ViewTop-OrgTop) / ZoomValue);
end;

procedure TNkVMainForm.GetColorValue(Ix, Iy: Integer; var R, G, B: Integer);
begin
  if (Ix<Image.Width) and (Iy<Image.Height) and (Ix>=0) and (Iy>=0) then
    Image.GetPixel(Ix, Iy, R, G, B)
  else begin
    R := -1;
    G := -1;
    B := -1;
  end;
end;

procedure TNkVMainForm.UpdateCaption;
var
  fname, info, mode: string;
  pt: TPoint;
  r, g, b, x, y, ix, iy: Integer;
begin
  if not Image.Loaded then
    Exit;

  if FileList.Items[FileList.Cur] = '' then
    fname := '<STDIN>'
  else
    fname := BaseName(FileList.Items[FileList.Cur]);

  if ArFileList.Count > 0 then
    fname := fname + '/' + ArFileList.Items[ArFileList.Cur];

  if ShowColorValue then
  begin
    GetCursorPos(pt);
    ScreenToClient(pt.x, pt.y, x, y);
    GetImageCoord(x, y, ix, iy);
    GetColorValue(ix, iy, r, g, b);
    if r > -1 then
      info := Fmt(
        '[R{0:03} G{1:03} B{2:03} #{0:02x}{1:02x}{2:02x}] (X{3} Y{4})',
        [r, g, b, ix, iy])
    else
      info := '[R--- G--- B---] (X--- Y---)';
  end else
    info := Fmt('({0}x{1}x{2})', [Image.Width, Image.Height, Image.Bpp]);

  if FitImage then
    mode := ' (Fit Image Mode)'
  else if FitLargeImage then
    mode := ' (Fit Large Image Mode)'
  else
    mode := '';

  SetText(Fmt('{0}  {1} {2}%{3} - {4}', [fname, info,
    IStr(RoundOff(ZoomValue*100)), mode, App.Name]));
end;

procedure TNkVMainForm.Error(const Path: string; const Msg: string);
var
  s: string;
begin
  if Path = '' then
    s := Msg
  else
    s := Path + EOL + EOL + Msg;
  MessageBox(Handle, PChar(s), PChar(App.Name), MB_ICONEXCLAMATION);
end;

procedure TNkVMainForm.SetWallpaper(Tile: Boolean);
var
  key: HKEY;
  data: array[0..MAX_PATH] of Char;
  fname: string;
begin
  if not Image.Loaded then
    Exit;

  fname := JoinPath([App.ProfDir, 'wallpaper.bmp']);
  try
    Image.Write(fname, Image.Width, Image.Height, False);
    if Tile then
      data := '1'
    else
      data := '0';
    if RegOpenKeyEx(HKEY_CURRENT_USER, 'Control Panel\desktop', 0, KEY_WRITE,
      key) = ERROR_SUCCESS then
    begin
      RegSetValueEx(key, 'TileWallPaper', 0, REG_SZ, @data, 1);
      RegCloseKey(key);
    end;
    SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, PChar(fname),
      SPIF_UPDATEINIFILE or SPIF_SENDWININICHANGE);
  except
    on E: EIOError do Error(E.Path, E.Msg);
  end;
end;

procedure TNkVMainForm.ActivateNextForm(Prev: Boolean);
var
  i, cur: Integer;
  wnds: HWNDArray;
begin
  wnds := ListWindow(ClassName);
  SortInt(TLongintArray(wnds));
  cur := 0;
  for i := 0 to High(wnds) do
  begin
    if wnds[i] = Handle then
    begin
      cur := i;
      Break;
    end;
  end;
  if Prev then
  begin
    Dec(cur);
    if cur < 0 then
      cur := High(wnds);
  end else
  begin
    Inc(cur);
    if cur = Length(wnds) then
      cur := 0;
  end;
  SetForegroundWindow(wnds[cur]);
end;

procedure TNkVMainForm.UpdateDispSize;
var
  cw, ch: Integer;
begin
  GetClientSize(cw, ch);
  if FitImage or
     (FitLargeImage and ((Image.Width > cw) or (Image.Height > ch))) then
  begin
    ViewLeft := 0;
    ViewTop := 0;
    if Image.Width * ch > Image.Height * cw then
    begin
      DispWidth := cw;
      DispHeight := RoundOff(Image.Height * DispWidth / Image.Width);
    end else
    begin
      DispHeight := ch;
      DispWidth := RoundOff(Image.Width * DispHeight / Image.Height);
    end;
  end
  else if FitLargeImage then
  begin
    ViewLeft := 0;
    ViewTop := 0;
    DispWidth := Image.Width;
    DispHeight := Image.Height;
  end;

  if DispWidth < cw then
    ViewWidth := DispWidth 
  else
    ViewWidth := cw;
  if DispHeight < ch then
    ViewHeight := DispHeight 
  else
    ViewHeight := ch;

  if ViewLeft > (DispWidth-ViewWidth) then
    ViewLeft := DispWidth-ViewWidth;
  if ViewTop > (DispHeight-ViewHeight) then
    ViewTop := DispHeight-ViewHeight;

  if ViewWidth < cw then
    OrgLeft := (cw - ViewWidth) div 2
  else
    OrgLeft := 0;
  if ViewHeight < ch then
    OrgTop := (ch - ViewHeight) div 2
  else
    OrgTop := 0;

  ZoomValue := DispWidth / Image.Width;

  UpdateCaption;
  Invalidate;
end;

{ Command Methods }

procedure TNkVMainForm.cmStayOnTop;
begin
  SetStayOnTop(not StayOnTop)
end;

procedure TNkVMainForm.cmHideBorder;
begin
  if not FullScreen then 
    SetBorder(not Border);
  UpdateCursor;
end;
  
procedure TNkVMainForm.cmFitImage;
begin
  FitImage := not FitImage;
  if FitImage and FitLargeImage then
    FitLargeImage := False;
  UpdateDispSize;
end;

procedure TNkVMainForm.cmFitLargeImage;
begin
  FitLargeImage := not FitLargeImage;
  if FitImage and FitLargeImage then
    FitImage := False;
  UpdateDispSize;
end;

procedure TNkVMainForm.cmScaleFilter;
begin
  ScaleFilter := not ScaleFilter;
  Invalidate;
end;

procedure TNkVMainForm.cmShowGrid;
begin
  ShowGrid := not ShowGrid;
  Invalidate;
end;

procedure TNkVMainForm.cmNextWindow;
begin
  ActivateNextForm(False);
end;

procedure TNkVMainForm.cmPrevWindow;
begin
  ActivateNextForm(True);
end;

procedure TNkVMainForm.cmAboutPlugins;
var
  i: Integer;
  s: string;
begin
  s := '';
  for i := 0 to High(Plugins) do
    with TSpiFilter.CreateByFile(Plugins[i]) do
      if Handle = 0 then
        s := s + '(not found)' + EOL
      else
        s := s + FileName + ': ' + Name + EOL;
  MsgBox(Handle, App.Name + ' - About Plugins', s);
end;

procedure TNkVMainForm.cmAbout;
begin
  App.About(
    App.Name+' '+Version+' ('+ReleaseDate+')'+EOL+EOL+
    Copyright+EOL+EOL+
    'This software is provided under the zlib/libpng license.'+EOL+EOL+
    'Profile Folder: '+App.ProfDir);
end;

procedure TNkVMainForm.cmExit;
begin
  Close;
end;

procedure TNkVMainForm.cmFitWindow;
begin
  if Image.Loaded and not FullScreen then
  begin
    if Maximized then
      Restore;
    SetClientSize(DispWidth, DispHeight);
  end;
end;

procedure TNkVMainForm.cmFullScreen;
begin
  if Image.Loaded then
    SetFullScreen(not FullScreen, False);
end;

procedure TNkVMainForm.cmZoomIn;
var
  i: Integer;
begin
  if Image.Loaded then
    for i := 0 to High(ZoomTable) do
      if DispWidth < RoundOff(Image.Width * ZoomTable[i]) then
      begin
        Zoom(ZoomTable[i]);
        Break;
      end;
end;

procedure TNkVMainForm.cmZoomOut;
var
  i: Integer;
begin
  if Image.Loaded then
    for i := High(ZoomTable) downto 0 do
      if DispWidth > RoundOff(Image.Width * ZoomTable[i]) then
      begin
        Zoom(ZoomTable[i]);
        Break;
      end;
end;

procedure TNkVMainForm.cmZoom50; begin Zoom(0.5); end;
procedure TNkVMainForm.cmZoom100; begin Zoom(1.0); end;
procedure TNkVMainForm.cmZoom150; begin Zoom(1.5); end;
procedure TNkVMainForm.cmZoom200; begin Zoom(2.0); end;
procedure TNkVMainForm.cmZoom300; begin Zoom(3.0); end;
procedure TNkVMainForm.cmZoom400; begin Zoom(4.0); end;
procedure TNkVMainForm.cmZoom800; begin Zoom(8.0); end;
  
procedure TNkVMainForm.cmSetWallpaper;
begin
  SetWallpaper(False);
end;

procedure TNkVMainForm.cmSetWallpaperTile;
begin
  SetWallpaper(True);
end;

procedure TNkVMainForm.cmScrollLeft;   begin Scroll(-KeyScroll, 0) end;
procedure TNkVMainForm.cmScrollRight;  begin Scroll(KeyScroll, 0) end;
procedure TNkVMainForm.cmScrollUp;     begin Scroll(0, -KeyScroll) end;
procedure TNkVMainForm.cmScrollDown;   begin Scroll(0, KeyScroll) end;
procedure TNkVMainForm.cmScrollTop;    begin Scroll(0, -DispHeight) end;
procedure TNkVMainForm.cmScrollBottom; begin Scroll(0, DispHeight) end;

procedure TNkVMainForm.cmScrollTopBottom;
var
  w, h: Integer;
begin
  GetClientSize(w, h);
  if Viewtop = DispHeight - h then
    Scroll(0, -DispHeight)
  else
    Scroll(0, DispHeight);
end;

procedure TNkVMainForm.cmShowColorValue;
begin
  if Image.Loaded then
  begin
    ShowColorValue := not ShowColorValue;
    UpdateCaption;
    UpdateCursor;
  end;
end;

procedure TNkVMainForm.cmFirstFile;
begin
  if Image.Loaded then
  begin
    FillFileList;
    FileList.Cur := -1;
    NextImage(FileList);
  end;
end;

procedure TNkVMainForm.cmLastFile;
begin
  if Image.Loaded then
  begin
    FillFileList;
    FileList.Cur := FileList.Count;
    PrevImage(FileList, False);
  end;
end;

procedure TNkVMainForm.cmNextFile;
begin
  if Image.Loaded then
  begin
    FillFileList;
    NextImage(FileList);
  end;
end;

procedure TNkVMainForm.cmPrevFile;
begin
  if Image.Loaded then
  begin
    FillFileList;
    PrevImage(FileList, False);
  end;
end;

procedure TNkVMainForm.cmNextImage;
begin
  if Image.Loaded then
    if (ArFileList.Count = 0) or (ArFileList.Cur = ArFilelist.Count-1) then
    begin
      FillFileList;
      NextImage(FileList)
    end else
      NextImage(ArFileList);
end;

procedure TNkVMainForm.cmPrevImage;
begin
  if Image.Loaded then
    if (ArFileList.Count = 0) or (ArFileList.Cur = 0) then
    begin
      FillFileList;
      PrevImage(FileList, True)
    end else
      PrevImage(ArFileList, True);
end;

procedure TNkVMainForm.cmCopyFileName;
begin
  if Image.Loaded then
    SetClipboardText(FileList.Items[FileList.Cur]);
end;

procedure TNkVMainForm.cmCopyColorValue;
var
  r, g, b, x, y, ix, iy: Integer;
  pt: TPoint;
begin
  if Image.Loaded then
  begin
    GetCursorPos(pt);
    ScreenToClient(pt.x, pt.y, x, y);
    GetImageCoord(x, y, ix, iy);
    GetColorValue(ix, iy, r, g, b);
    if r > -1 then
      SetClipboardText(Fmt('#{0:02x}{1:02x}{2:02x}', [r, g, b]));
  end;
end;

procedure TNkVMainForm.cmSaveAs;
var
  fname: string;
begin
  if not Image.Loaded then
    Exit;

  if FileList.Items[FileList.Cur] = '' then
    fname := 'stdin.bmp'
  else
    fname := BaseName(FileList.Items[FileList.Cur]) + '.bmp';
  if ShowSaveDlg(Handle, ['All Files', '*.*'], fname) then
    try
      Image.Write(fname, DispWidth, DispHeight, ScaleFilter);
    except
      on E: EIOError do Error(E.Path, E.Msg);
    end;
end;

procedure TNkVMainForm.cmInfo;
begin
  if Image.Loaded then
  begin
    UpdateInfo;
    if not IsWindow(InfoDlg.Handle) then
      InfoDlg.ShowModeless;
  end;
end;

procedure TNkVMainForm.cmReloadImage;
begin
  if Image.Loaded then
  begin
    LoadFile(FileList.Items[FileList.Cur], False);
    UpdateDispSize;
    UpdateWinSize;
    UpdateInfo;
  end;
end;

procedure TNkVMainForm.cmRefreshFileList;
begin
  if Image.Loaded and DirMode then
    FileListUpdate := True; 
end;

procedure TNkVMainForm.cmPluginMenu; begin end; // dummy

procedure TNkVMainForm.cmSetImageSize;
begin
  if Image.Loaded then
    with TImageSizeDlg.CreateByResource(Handle, 'IMAGESIZEDIALOG') do
      try
        Image := Self.Image;
        Width := DispWidth;
        Height := DispHeight;
        ShowModal;
        if Update then
        begin
          FitImage := False;
          FitLargeImage := False;
          DispWidth := Width;
          DispHeight := Height;
          UpdateDispSize;
        end;
      finally
        Free;
      end;
end;

{ Message Handlers }

procedure TNkVMainForm.WMSize(var Message: TWMSize);
begin
  if Image.Loaded then
    UpdateDispSize;
end;

procedure TNkVMainForm.WMPaint(var Message: TWMPaint);
var
  dc: HDC;
  ps: TPaintStruct;
  cw, ch: Integer;
  br: HBRUSH;
begin
  dc := BeginPaint(Handle, ps);
  
  br := CreateSolidBrush(BgColor);
  GetClientSize(cw, ch);

  if Image.Loaded then
  begin
    Image.Draw(dc, OrgLeft, OrgTop, ViewWidth, ViewHeight, ViewLeft, ViewTop,
      DispWidth, DispHeight, ShowGrid and (DispWidth div Image.Width >= 1),
      ScaleFilter);

    if ViewWidth < cw then
    begin
      // right
      FillRect(dc, Rect(OrgLeft+ViewWidth, OrgTop, cw, OrgTop+ViewHeight), br);
      // left
      FillRect(dc, Rect(0, OrgTop, OrgLeft, OrgTop+ViewHeight), br);
    end;
    if ViewHeight < ch then
    begin
      FillRect(dc, Rect(0, OrgTop+ViewHeight, cw, ch), br); // bottom
      FillRect(dc, Rect(0, 0, cw, OrgTop), br); // top
    end;
  end else
    FillRect(dc, Rect(0, 0, cw, ch), br);

  DeleteObject(br); 
  EndPaint(Handle, ps);
end;

procedure TNkVMainForm.ShowPluginConfig(const FileName: string);
begin
  with TSpiFilter.CreateByFile(FileName) do
    try
      if Handle = 0 then
      begin
        Error(FileName, 'Load failed.');
        Exit;
      end;
      try
        ShowConfigDlg(Self.Handle);
      except
        on E: EImgFilterError do Error(FileName, E.Msg);
      end;
    finally
      Free;
    end;
end;

procedure TNkVMainForm.WMCommand(var Message: TWMCommand);
begin
  case Message.ItemID of
    IDM_PLUGINSETTING_FIRST .. IDM_PLUGINSETTING_LAST:
      ShowPluginConfig(Plugins[Message.ItemID - IDM_PLUGINSETTING_FIRST]);
  else
    Command.DispatchCommand(Message.ItemID);
  end;
end;

procedure TNkVMainForm.WMLButtonDown(var Message: TWMLButtonDown);
begin
  if Border or FullScreen or (not Border and (GetKeyState(VK_SHIFT) < 0)) then
  begin
    SetCapture(Handle);
    if Image.Loaded then
      SetCursor(GrabCursor);
    GetCursorPos(OldPt);
  end else
  begin
    ReleaseCapture;
    SendMsg(WM_NCLBUTTONDOWN, HTCAPTION, 0);
  end;
end;

procedure TNkVMainForm.WMLButtonUp(var Message: TWMLButtonUp);
begin
  ReleaseCapture;
  //SetCursor(NormalCursor);
end;

procedure TNkVMainForm.WMMouseMove(var Message: TWMMouseMove);
var
  pt: TPoint;
begin
  if Message.Keys and MK_LBUTTON <> 0 then
  begin
    // If mouse cursor are on outside window,
    // coordinates included in Message are invalid. So use GetCursorPos()
    GetCursorPos(pt);
    Scroll(OldPt.x - pt.x, OldPt.y - pt.y);
    OldPt := pt;
  end
  else if ShowColorValue then
    UpdateCaption;
end;

procedure TNkVMainForm.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  cmFitWindow;
end;

procedure TNkVMainForm.WMInitMenuPopup(var Message: TMessage);
begin
  if Assigned(ContextMenu) then
    ContextMenu.HandleMenuMsg(Message);
end;

procedure TNkVMainForm.WMDrawItem(var Message: TMessage);
begin
  if Assigned(ContextMenu) then
    ContextMenu.HandleMenuMsg(Message);
end;

procedure TNkVMainForm.WMMeasureItem(var Message: TMessage);
begin
  if Assigned(ContextMenu) then
    ContextMenu.HandleMenuMsg(Message);
end;

procedure TNkVMainForm.WMContextMenu(var Message: TMessage);
var
  i, id, x, y: Integer;
  menu: TPopupMenu;
begin
  ScreenToClient(Message.LParamLo, Message.LParamHi, x, y);
  if y < 0 then
  begin
    inherited;
    Exit;
  end;
 
  menu := TPopupMenu.Create;
  
  if GetKeyState(VK_SHIFT) < 0 then
  begin
    if Image.Loaded and (FileList.Items[FileList.Cur] <> '') then
    begin
      ContextMenu := TContextMenu.Create(Handle, menu.Handle,
        FileList.Items[FileList.Cur]);
      if Message.LParam = -1 then
        ClientToScreen(0, 0, x, y)
      else begin
        x := Message.LParamLo;
        y := Message.LParamHi;
      end;
      id := Integer(TrackPopupMenu(menu.Handle,
        TPM_LEFTALIGN or TPM_RIGHTBUTTON or TPM_RETURNCMD,
        x, y, 0, Handle, nil));
      ContextMenu.Invoke(id);
      ContextMenu.Free;
      ContextMenu := nil;
      menu.Free;
    end;
    Exit;
  end;

  with menu do
    try
      Command.AddMenuItems(Handle, MenuItems);
      if Image.Loaded then
      begin
        if FileList.Items[FileList.Cur] = '' then
        begin
          GrayItem(Command.GetId(cmPrevImage));
          GrayItem(Command.GetId(cmNextImage));
          GrayItem(Command.GetId(cmPrevFile));
          GrayItem(Command.GetId(cmNextFile));
          GrayItem(Command.GetId(cmCopyFileName));
          GrayItem(Command.GetId(cmReloadImage));
          GrayItem(Command.GetId(cmLastFile));
          GrayItem(Command.GetId(cmFirstFile));
        end;
      end else
      begin
        GrayItem(Command.GetId(cmPrevImage));
        GrayItem(Command.GetId(cmNextImage));
        GrayItem(Command.GetId(cmPrevFile));
        GrayItem(Command.GetId(cmNextFile));
        GrayItem(Command.GetId(cmFirstFile));
        GrayItem(Command.GetId(cmLastFile));
        GrayItem(Command.GetId(cmInfo));
        GrayItem(Command.GetId(cmFullScreen));
        GrayItem(Command.GetId(cmFitWindow));
        GrayItem(Command.GetId(cmZoom50));
        GrayItem(Command.GetId(cmZoom100));
        GrayItem(Command.GetId(cmZoom150));
        GrayItem(Command.GetId(cmZoom200));
        GrayItem(Command.GetId(cmZoom300));
        GrayItem(Command.GetId(cmZoom400));
        GrayItem(Command.GetId(cmZoom800));
        GrayItem(Command.GetId(cmZoomIn));
        GrayItem(Command.GetId(cmZoomOut));
        GrayItem(Command.GetId(cmShowColorValue));
        GrayItem(Command.GetId(cmCopyFileName));
        GrayItem(Command.GetId(cmReloadImage));
        GrayItem(Command.GetId(cmSaveAs));
        GrayItem(Command.GetId(cmSetWallpaper));
        GrayItem(Command.GetId(cmSetWallpaperTile));
        GrayItem(Command.GetId(cmSetImageSize));
      end; 

      if FullScreen then
      begin
        GrayItem(Command.GetId(cmFitWindow));
        GrayItem(Command.GetId(cmHideBorder));
      end;

      id := Command.GetId(cmPluginMenu);
      for i := 0 to High(Plugins) do
        InsertItem(id, BaseName(Plugins[i]), IDM_PLUGINSETTING_FIRST+i);
      DeleteItem(id);

      CheckItem(Command.GetId(cmFullScreen), FullScreen);
      CheckItem(Command.GetId(cmStayOnTop), StayOnTop);
      CheckItem(Command.GetId(cmFitImage), FitImage);
      CheckItem(Command.GetId(cmFitLargeImage), FitLargeImage);
      CheckItem(Command.GetId(cmHideBorder), not Border);
      CheckItem(Command.GetId(cmShowColorValue), ShowColorValue);
      CheckItem(Command.GetId(cmScaleFilter), ScaleFilter);
      CheckItem(Command.GetId(cmShowGrid), ShowGrid);
      if Message.LParam = -1 then
      begin
        ClientToScreen(0, 0, x, y);
        Popup(Self.Handle, x, y);
      end else
        Popup(Self.Handle, Message.LParamLo, Message.LParamHi);
    finally
      Free;
    end;
end;

procedure TNkVMainForm.WMDropFiles(var Message: TWMDropFiles);
var
  n, i: Integer;
  path: string;
  list: TFileList;
  dir: TDir;
begin
  list := TFileList.Create;
  n := DragQueryFile(Message.Drop, $FFFFFFFF, nil, 0);
  for i := 0 to n-1 do
  begin
    SetLength(path, DragQueryFile(Message.Drop, i, nil, 0));
    // +1 for null terminate
    DragQueryFile(Message.Drop, i, PChar(path), Length(path)+1);
    if IsDir(path) then
    begin
      DirMode := n = 1;
      with TListDir.Create(path) do
        try
          while Each(dir) do
            if not (dmDirectory in dir.Mode) then
              list.Add(JoinPath([path, dir.Name]));
        finally
          Free;
        end
    end else
      list.Add(path);
  end;
  DragFinish(Message.Drop);
  list.Sort;
  LoadFirstFile(list);
  list.Free;
  if (FileList.Count = 1) and (FileList.Items[0] <> '') then // not stdin 
  begin
    DirMode := True;
    FileListUpdate := True;
  end;
end;

procedure TNkVMainForm.WMMouseWheel(var Message: TMessage);
begin
  if Message.WParam > 0 then
    cmPrevImage
  else
    cmNextImage;
end;

procedure TNkVMainForm.WMClose(var Message: TWMClose);
begin
  SavePos;
  inherited;
end;

procedure TNkVMainForm.WMQueryEndSession(var Message: TWMQueryEndSession);
begin
  SavePos;
  Message.Result := Integer(True);
end;

end.
